"""Tests for the Bulb API with a light strip."""
from typing import AsyncGenerator

import pytest

from pywizlight import PilotBuilder, wizlight
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def light_strip() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP20_SHRGB_01ABI", firmware_version="1.25.0"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_setting_rgbww(light_strip: wizlight) -> None:
    """Test setting rgbww."""
    await light_strip.turn_on(PilotBuilder(rgbww=(1, 2, 3, 4, 5)))
    state = await light_strip.updateState()
    assert state and state.get_rgbww() == (1, 2, 3, 4, 5)


@pytest.mark.asyncio
async def test_supported_scenes(light_strip: wizlight) -> None:
    """Test supported scenes."""
    assert await light_strip.getSupportedScenes() == [
        "Ocean",
        "Romance",
        "Sunset",
        "Party",
        "Fireplace",
        "Cozy",
        "Forest",
        "Pastel Colors",
        "Wake up",
        "Bedtime",
        "Warm White",
        "Daylight",
        "Cool white",
        "Night light",
        "Focus",
        "Relax",
        "True colors",
        "TV time",
        "Plantgrowth",
        "Spring",
        "Summer",
        "Fall",
        "Deepdive",
        "Jungle",
        "Mojito",
        "Club",
        "Christmas",
        "Halloween",
        "Candlelight",
        "Golden white",
        "Pulse",
        "Steampunk",
        "Rhythm",
    ]


@pytest.mark.asyncio
async def test_model_description_light_strip(light_strip: wizlight) -> None:
    """Test fetching the model description for a light strip."""
    bulb_type = await light_strip.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=True, color_tmp=True, effect=True, brightness=True, dual_head=False
        ),
        name="ESP20_SHRGB_01ABI",
        kelvin_range=KelvinRange(max=6500, min=2700),
        bulb_type=BulbClass.RGB,
        fw_version="1.25.0",
        white_channels=2,
        white_to_color_ratio=80,
    )



"""Start up a fake bulb to test features without a real bulb."""
import asyncio
import json
from typing import Any, Callable, Dict, Tuple, cast

from pywizlight.protocol import WizProtocol

MODULE_CONFIGS = {  # AKA getModelConfig
    ("ESP01_SHRGB_03", "1.25.0"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 1,
            "pwmFreq": 1000,
            "pwmRange": [3, 100],
            "wcr": 30,
            "nowc": 1,
            "cctRange": [2200, 2700, 4800, 6500],
            "renderFactor": [171, 255, 75, 255, 43, 85, 0, 0, 0, 0],
        },
    },
    ("ESP10_SOCKET_06", "1.25.0"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 1,
            "pwmFreq": 200,
            "pwmRange": [1, 100],
            "wcr": 20,
            "nowc": 2,
            "cctRange": [2700, 2700, 2700, 2700],
            "renderFactor": [255, 0, 255, 255, 0, 0, 0, 0, 0, 0],
        },
    },
    ("ESP05_SHDW_21", "1.25.0"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 1,
            "pwmFreq": 1000,
            "pwmRange": [5, 100],
            "wcr": 20,
            "nowc": 1,
            "cctRange": [2700, 2700, 2700, 2700],
            "renderFactor": [255, 0, 255, 255, 0, 0, 0, 0, 0, 0],
        },
    },
    ("ESP20_SHRGB_01ABI", "1.25.0"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 0,
            "pwmFreq": 2000,
            "pwmRange": [5, 100],
            "wcr": 80,
            "nowc": 2,
            "cctRange": [2700, 2700, 6500, 6500],
            "renderFactor": [255, 0, 255, 255, 0, 0, 0, 81, 245, 178],
            "drvIface": 0,
        },
    },
    ("ESP21_SHTW_01", "1.25.0"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 2,
            "pwmFreq": 5000,
            "pwmRange": [1, 100],
            "wcr": 20,
            "nowc": 1,
            "cctRange": [2700, 2700, 5000, 5000],
            "renderFactor": [255, 0, 255, 255, 0, 0, 0, 0, 0, 0],
            "drvIface": 0,
        },
    },
    ("MISSING", "1.16.64"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 2,
            "pwmFreq": 5000,
            "pwmRange": [1, 100],
            "wcr": 20,
            "nowc": 1,
            "cctRange": [2700, 2700, 5000, 5000],
            "renderFactor": [255, 0, 255, 255, 0, 0, 0, 0, 0, 0],
            "drvIface": 0,
        },
    },
    ("MISSING_KELVIN", "1.16.64"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 2,
            "pwmFreq": 5000,
            "pwmRange": [1, 100],
            "wcr": 20,
            "nowc": 1,
            "renderFactor": [255, 0, 255, 255, 0, 0, 0, 0, 0, 0],
            "drvIface": 0,
        },
    },
    ("INVALID", "1.16.64"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 2,
            "pwmFreq": 5000,
            "pwmRange": [1, 100],
            "wcr": 20,
            "nowc": 1,
            "cctRange": [2700, 2700, 5000, 5000],
            "renderFactor": [255, 0, 255, 255, 0, 0, 0, 0, 0, 0],
            "drvIface": 0,
        },
    },
    ("ESP20_SHRGB_01BT", "1.23.70"): {
        "method": "getModelConfig",
        "env": "pro",
        "result": {
            "ps": 1,
            "pwmFreq": 2000,
            "pwmRange": [0, 100],
            "wcr": 20,
            "nowc": 2,
            "cctRange": [2200, 2700, 6500, 6500],
            "renderFactor": [200, 255, 255, 150, 0, 0, 40, 0, 0, 0],
            "hasAdjMinDim": 0,
            "hasTapSensor": 1,
            "i2cTapSensor": [
                {
                    "chip": "SC7A22",
                    "whoami": [24, 25, 15, 19],
                    "freq": 100,
                    "shockWin": 40,
                },
                {
                    "chip": "IIS2DLPC",
                    "whoami": [24, 25, 15, 68],
                    "freq": 100,
                    "shockWin": 40,
                },
                {
                    "chip": "DA262",
                    "whoami": [38, 39, 1, 19],
                    "freq": 100,
                    "shockWin": 50,
                },
            ],
            "pm": 0,
            "fanSpeed": 0,
        },
    },
}

SYSTEM_CONFIGS: Dict[Tuple[str, str], Any] = {  # AKA getSystemConfig
    ("1.8.0-MISSING-TYPEID-0", "1.8.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb502054e3",
            "homeId": 5385975,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "fwVersion": "1.8.0",
            "groupId": 0,
            "drvConf": [20, 1],
        },
    },
    ("1.8.0-MISSING-TYPEID-1", "1.8.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb502054e3",
            "homeId": 5385975,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 1,
            "fwVersion": "1.8.0",
            "groupId": 0,
            "drvConf": [20, 1],
        },
    },
    ("ESP01_SHRGB_03", "1.25.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb5006033d",
            "homeId": 653906,
            "roomId": 989983,
            "moduleName": "ESP01_SHRGB_03",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "drvConf": [30, 1],
            "ewf": [255, 0, 255, 255, 0, 0, 0],
            "ewfHex": "ff00ffff000000",
            "ping": 0,
        },
    },
    ("ESP10_SOCKET_06", "1.25.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb5006033d",
            "homeId": 653906,
            "roomId": 989983,
            "moduleName": "ESP10_SOCKET_06",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "drvConf": [20, 2],
            "ewf": [255, 0, 255, 255, 0, 0, 0],
            "ewfHex": "ff00ffff000000",
            "ping": 0,
        },
    },
    ("ESP10_SOCKET_06", "1.16.71"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb50df7693",
            "homeId": 5385975,
            "roomId": 8201410,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "moduleName": "ESP10_SOCKET_06",
            "fwVersion": "1.16.71",
            "groupId": 0,
            "drvConf": [20, 2],
        },
    },
    ("ESP05_SHDW_21", "1.25.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb5006033d",
            "homeId": 653906,
            "roomId": 989983,
            "moduleName": "ESP05_SHDW_21",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "drvConf": [20, 1],
            "ewf": [255, 0, 255, 255, 0, 0, 0],
            "ewfHex": "ff00ffff000000",
            "ping": 0,
        },
    },
    ("ESP20_SHRGB_01ABI", "1.25.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb5006033d",
            "homeId": 653906,
            "roomId": 989983,
            "moduleName": "ESP20_SHRGB_01ABI",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "drvConf": [80, 2],
            "ewf": [255, 0, 255, 255, 0, 0, 0],
            "ewfHex": "ff00ffff000000",
            "ping": 0,
        },
    },
    ("ESP20_SHRGB_01ABI", "1.21.4"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb5006033d",
            "homeId": 653906,
            "roomId": 989983,
            "moduleName": "ESP20_SHRGB_01ABI",
            "fwVersion": "1.21.4",
            "groupId": 0,
            "drvConf": [80, 2],
            "ewf": [255, 0, 255, 255, 0, 0, 0],
            "ewfHex": "ff00ffff000000",
            "ping": 0,
        },
    },
    ("ESP03_SHRGB3_01ABI", "1.16.64"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb50fd92e5",
            "homeId": 5385975,
            "roomId": 0,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "moduleName": "ESP03_SHRGB3_01ABI",
            "fwVersion": "1.16.64",
            "groupId": 0,
            "drvConf": [20, 1],
        },
    },
    ("ESP20_SHRGBC_01", "1.21.4"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "6c2990e493bc",
            "homeId": 5385975,
            "roomId": 8016844,
            "moduleName": "ESP20_SHRGBC_01",
            "fwVersion": "1.21.4",
            "groupId": 0,
            "drvConf": [30, 1],
            "ewf": [200, 255, 150, 255, 0, 0, 40],
            "ewfHex": "c8ff96ff000028",
            "ping": 0,
        },
    },
    ("ESP21_SHTW_01", "1.25.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "6c29905a067c",
            "homeId": 5385975,
            "roomId": 8016844,
            "rgn": "eu",
            "moduleName": "ESP21_SHTW_01",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "ping": 0,
        },
    },
    ("MISSING_KELVIN", "1.16.64"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "6c29905a067c",
            "homeId": 5385975,
            "roomId": 8016844,
            "rgn": "eu",
            "moduleName": "ESP21_SHTW_01",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "ping": 0,
        },
    },
    ("MISSING", "1.16.64"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "6c29905a067c",
            "homeId": 5385975,
            "roomId": 8016844,
            "rgn": "eu",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "ping": 0,
        },
    },
    ("INVALID", "1.16.64"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "6c29905a067c",
            "homeId": 5385975,
            "roomId": 8016844,
            "rgn": "eu",
            "moduleName": "INVALID",
            "fwVersion": "1.25.0",
            "groupId": 0,
            "ping": 0,
        },
    },
    ("ESP01_SHRGB1C_31", "1.17.1"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb50bdf8d7",
            "homeId": 5385975,
            "roomId": 0,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "moduleName": "ESP01_SHRGB1C_31",
            "fwVersion": "1.17.1",
            "groupId": 0,
            "drvConf": [20, 2],
        },
    },
    ("ESP14_SHTW1C_01", "1.18.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb503ea5f4",
            "homeId": 5385975,
            "roomId": 0,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "moduleName": "ESP14_SHTW1C_01",
            "fwVersion": "1.18.0",
            "groupId": 0,
            "drvConf": [20, 1],
        },
    },
    ("ESP05_SHTW_21", "1.10.0"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb5007bfc6",
            "homeId": 5385975,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "moduleName": "ESP05_SHTW_21",
            "fwVersion": "1.10.0",
            "groupId": 0,
            "drvConf": [20, 1],
        },
    },
    ("ESP06_SHDW9_01", "1.11.7"): {
        "method": "getSystemConfig",
        "env": "",
        "result": {
            "mac": "a8bb509f71d1",
            "homeId": 0,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "moduleName": "ESP06_SHDW9_01",
            "fwVersion": "1.11.7",
            "groupId": 0,
            "drvConf": [20, 1],
        },
    },
    ("ESP20_DHRGB_01B", "1.21.40"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "d8a01199cf31",
            "homeId": 5385975,
            "roomId": 8201410,
            "moduleName": "ESP20_DHRGB_01B",
            "fwVersion": "1.21.40",
            "groupId": 0,
            "drvConf": [20, 2],
            "ewf": [200, 255, 150, 255, 0, 0, 40],
            "ewfHex": "c8ff96ff000028",
            "ping": 0,
        },
    },
    ("ESP20_SHRGB_01BT", "1.23.70"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "d8a0119c42df",
            "homeId": 5385975,
            "roomId": 0,
            "rgn": "eu",
            "moduleName": "ESP20_SHRGB_01BT",
            "fwVersion": "1.23.70",
            "groupId": 0,
            "ping": 0,
        },
    },
    ("BROKEN_JSON", "1.0.0"): json.JSONDecodeError,
    ("ESP01_DIMTRIACS_01", "1.16.68"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "a8bb50f30985",
            "homeId": 5385975,
            "roomId": 8201410,
            "homeLock": False,
            "pairingLock": False,
            "typeId": 0,
            "moduleName": "ESP01_DIMTRIACS_01",
            "fwVersion": "1.16.68",
            "groupId": 0,
            "drvConf": [20, 1],
        },
    },
    ("ESP25_SOCKET_01", "1.26.2"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "d8a0119906b7",
            "homeId": "**REDACTED**",
            "roomId": "**REDACTED**",
            "rgn": "eu",
            "moduleName": "ESP25_SOCKET_01",
            "fwVersion": "1.26.2",
            "groupId": 0,
            "ping": 0,
        },
    },
    ("ESP25_SOCKET_01", "1.26.1"): {
        "method": "getSystemConfig",
        "env": "pro",
        "result": {
            "mac": "d8a0119906b7",
            "homeId": "**REDACTED**",
            "roomId": "**REDACTED**",
            "rgn": "eu",
            "moduleName": "ESP25_SOCKET_01",
            "fwVersion": "1.26.2",
            "groupId": 0,
            "ping": 0,
        },
    },
}

USER_CONFIGS: Dict[Tuple[str, str], Any] = {  # AKA getUserConfig
    ("ESP20_SHRGB_01ABI", "1.21.4"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 450,
            "fadeOut": 500,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2700, 6500],
            "opMode": 0,
            "po": False,
        },
    },
    ("ESP03_SHRGB3_01ABI", "1.16.64"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 450,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2700, 6500],
            "opMode": 0,
            "po": False,
        },
    },
    ("ESP20_SHRGBC_01", "1.21.4"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 500,
            "fadeOut": 500,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2200, 6500],
            "opMode": 0,
            "po": True,
        },
    },
    ("ESP01_SHRGB1C_31", "1.17.1"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 450,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2700, 6500],
            "po": False,
        },
    },
    ("ESP05_SHTW_21", "1.10.0"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 450,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2700, 6500],
        },
    },
    ("ESP14_SHTW1C_01", "1.18.0"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 450,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2700, 6500],
            "opMode": 0,
            "po": False,
        },
    },
    ("ESP06_SHDW9_01", "1.11.7"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 450,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2700, 6500],
        },
    },
    ("ESP20_DHRGB_01B", "1.21.40"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 500,
            "fadeOut": 500,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2200, 6500],
            "opMode": 0,
            "po": True,
        },
    },
    ("1.8.0-MISSING-TYPEID-0", "1.8.0"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 500,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [5, 100],
            "whiteRange": [2700, 2700],
        },
    },
    ("1.8.0-MISSING-TYPEID-1", "1.8.0"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 500,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [5, 100],
            "whiteRange": [2700, 2700],
        },
    },
    ("BROKEN_JSON", "1.0.0"): json.JSONDecodeError,
    ("ESP10_SOCKET_06", "1.16.71"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 450,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [0, 100],
            "whiteRange": [2700, 6500],
            "extRange": [2700, 6500],
            "po": False,
        },
    },
    ("ESP01_DIMTRIACS_01", "1.16.68"): {
        "method": "getUserConfig",
        "env": "pro",
        "result": {
            "fadeIn": 500,
            "fadeOut": 500,
            "fadeNight": False,
            "dftDim": 100,
            "pwmRange": [10, 100],
            "whiteRange": [2700, 2700],
            "extRange": [2700, 2700],
            "po": True,
        },
    },
}

MODEL_CONFIG_NOT_FOUND = {
    "method": "getModelConfig",
    "env": "pro",
    "error": {"code": -32601, "message": "Method not found"},
}


USER_CONFIG_NOT_FOUND = {
    "method": "getUserConfig",
    "env": "pro",
    "error": {"code": -32601, "message": "Method not found"},
}

GET_POWER_NOT_FOUND = {
    "method": "getPower",
    "env": "pro",
    "error": {"code": -32601, "message": "Method not found"},
}


def get_initial_pilot() -> Dict[str, Any]:
    return {
        "method": "getPilot",
        "env": "pro",
        "result": {
            "mac": "ABCABCABCABC",
            "rssi": -62,
            "src": "udp",
            "state": False,
            "sceneId": 0,
            "r": 255,
            "g": 127,
            "b": 0,
            "c": 0,
            "w": 0,
            "temp": 0,
            "dimming": 13,
        },
    }


def get_initial_sys_config(module_name: str, firmware_version: str) -> Dict[str, Any]:
    return SYSTEM_CONFIGS[(module_name, firmware_version)]


def get_initial_model_config(module_name: str, firmware_version: str) -> Dict[str, Any]:
    return MODULE_CONFIGS.get(
        (module_name, firmware_version),
        MODEL_CONFIG_NOT_FOUND,
    )


def get_initial_user_config(module_name: str, firmware_version: str) -> Dict[str, Any]:
    return USER_CONFIGS.get((module_name, firmware_version), USER_CONFIG_NOT_FOUND)


def get_power(module_name: str, firmware_version: str) -> Dict[str, Any]:
    if module_name == "ESP25_SOCKET_01" and firmware_version == "1.26.2":
        return {"method": "getPower", "env": "pro", "result": {"power": 1065385}}
    if module_name == "ESP25_SOCKET_01" and firmware_version == "1.26.1":
        return {
            "method": "getPower",
            "env": "pro",
        }
    return GET_POWER_NOT_FOUND


BULB_JSON_ERROR = b'{"env":"pro","error":{"code":-32700,"message":"Parse error"}}'


class BulbUDPRequestHandler:
    """Class for UDP handler."""

    pilot_state: Dict[str, Any]  # Will be set by constructor for the actual class
    sys_config: Dict[str, Any]  # Will be set by constructor for the actual class
    model_config: Dict[str, Any]  # Will be set by constructor for the actual class
    user_config: Dict[str, Any]
    registration: Dict[str, Any]
    get_power: Dict[str, Any]
    transport: asyncio.DatagramTransport

    def handle(self, resp: bytes, addr: Tuple[str, int]) -> None:
        """Handle the request."""
        data = resp.strip()
        print(f"Request:{data!r}")
        try:
            json_data: Dict[str, Any] = dict(json.loads(data.decode()))
        except json.JSONDecodeError:
            self.transport.sendto(BULB_JSON_ERROR, addr)
            return

        method = str(json_data["method"])
        if method == "setPilot":
            return_data = self.setPilot(json_data)
            self.transport.sendto(return_data, addr)
        elif method == "getPilot":
            print(f"Response:{json.dumps(self.pilot_state)!r}")
            self.transport.sendto(bytes(json.dumps(self.pilot_state), "utf-8"), addr)
        elif method == "getSystemConfig":
            if self.sys_config == json.JSONDecodeError:
                self.transport.sendto(b"garbage", addr)
                self.transport.sendto(b"garbage", addr)
            else:
                self.transport.sendto(bytes(json.dumps(self.sys_config), "utf-8"), addr)
                # Simulate late response coming in twice
                self.transport.sendto(bytes(json.dumps(self.sys_config), "utf-8"), addr)
        elif method == "getModelConfig":
            self.transport.sendto(bytes(json.dumps(self.model_config), "utf-8"), addr)
        elif method == "getPower":
            self.transport.sendto(bytes(json.dumps(self.get_power), "utf-8"), addr)
        elif method == "getUserConfig":
            # Simulate late response of model config missing to ensure
            # it does not break getUserConfig
            self.transport.sendto(
                bytes(json.dumps(MODEL_CONFIG_NOT_FOUND), "utf-8"), addr
            )
            if self.user_config == json.JSONDecodeError:
                self.transport.sendto(b"garbage", addr)
            else:
                self.transport.sendto(
                    bytes(json.dumps(self.user_config), "utf-8"), addr
                )
        elif method == "registration":
            # Simulate late response of model config missing to ensure
            # it does not break registration
            self.transport.sendto(
                bytes(json.dumps(MODEL_CONFIG_NOT_FOUND), "utf-8"), addr
            )
            self.transport.sendto(bytes(json.dumps(self.registration), "utf-8"), addr)
        else:
            raise RuntimeError(f"No handler for {method}")

    def setPilot(self, json_data: Dict[str, Any]) -> bytes:
        """Change the values in the state."""
        for name, value in json_data["params"].items():
            self.pilot_state["result"][name] = value
        return b'{"method":"setPilot","env":"pro","result":{"success":true}}'


async def make_udp_fake_bulb_server(
    module_name: str, firmware_version: str
) -> Tuple[asyncio.BaseTransport, asyncio.BaseProtocol]:
    """Configure a fake bulb instance."""
    handler = BulbUDPRequestHandler()
    handler.pilot_state = get_initial_pilot()
    handler.sys_config = get_initial_sys_config(module_name, firmware_version)
    handler.model_config = get_initial_model_config(module_name, firmware_version)
    handler.user_config = get_initial_user_config(module_name, firmware_version)
    handler.registration = {
        "method": "registration",
        "env": "pro",
        "result": {"mac": "a8bb5006033d", "success": True},
    }
    handler.get_power = get_power(module_name, firmware_version)

    transport_proto = await asyncio.get_event_loop().create_datagram_endpoint(
        lambda: WizProtocol(on_response=handler.handle),
        local_addr=("127.0.0.1", 0),
    )
    handler.transport = cast(asyncio.DatagramTransport, transport_proto[0])
    return transport_proto


async def make_udp_fake_bulb_push_server() -> Tuple[
    asyncio.BaseTransport, asyncio.BaseProtocol
]:
    """Configure a fake push instance."""
    handler = BulbUDPRequestHandler()
    transport_proto = await asyncio.get_event_loop().create_datagram_endpoint(
        lambda: WizProtocol(on_response=lambda resp, addr: None),
        local_addr=("127.0.0.1", 0),
    )
    handler.transport = cast(asyncio.DatagramTransport, transport_proto[0])
    return transport_proto


async def startup_bulb(
    module_name: str = "ESP01_SHRGB_03", firmware_version: str = "1.25.0"
) -> Tuple[Callable[[], Any], int]:
    """Start up the bulb. Returns a function to shut it down."""
    transport_proto = await make_udp_fake_bulb_server(module_name, firmware_version)
    transport = cast(asyncio.DatagramTransport, transport_proto[0])
    return transport.close, transport.get_extra_info("sockname")[1]



"""Library with compatible bulb types.

Bulb Type detection:
ESP01_SHDW1C_31
ESP01 -- defines the module family (WiFi only bulb in this case)
DH -- Dual Head light
SH -- Single Head light (most bulbs are single heads) / LED Strip
TW -- Tunable White - can only control CCT and dimming; no color
DW -- Dimmable White (most filament bulbs)
RGB -- Fullstack bulb
1C -- Specific to the hardware - defines PWM frequency + way of controlling CCT temperature
31 -- Related to the hardware revision
"""
import dataclasses
import logging
from enum import Enum
from typing import List, Optional

from pywizlight.exceptions import WizLightNotKnownBulb

_LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class Features:
    """Defines the supported features."""

    color: bool
    color_tmp: bool
    effect: bool
    brightness: bool
    dual_head: bool


@dataclasses.dataclass(frozen=True)
class KelvinRange:
    """Defines the kelvin range."""

    max: int
    min: int


class BulbClass(Enum):
    """Bulb Types."""

    TW = "Tunable White"
    """Have Cool White and Warm White LEDs."""
    DW = "Dimmable White"
    """Have only Dimmable white LEDs."""
    RGB = "RGB Tunable"
    """Have RGB LEDs."""
    SOCKET = "Socket"
    """Smart socket with only on/off."""


KNOWN_TYPE_IDS = {0: BulbClass.DW}

_BASE_FEATURE_MAP = {
    # RGB supports effects and tuneable white
    BulbClass.RGB: {
        "brightness": True,
        "color": True,
        "color_tmp": True,
    },
    # TODO: TW supports effects but only "some"; improve the mapping to supported effects
    BulbClass.TW: {
        "brightness": True,
        "color": False,
        "color_tmp": True,
    },
    # Dimmable white only supports brightness and some basic effects
    BulbClass.DW: {
        "brightness": True,
        "color": False,
        "color_tmp": False,
    },
    # Socket supports only on/off
    BulbClass.SOCKET: {
        "brightness": False,
        "color": False,
        "color_tmp": False,
    },
}


@dataclasses.dataclass(frozen=True)
class BulbType:
    """BulbType object to define functions and features of the bulb."""

    features: Features
    name: Optional[str]
    kelvin_range: Optional[KelvinRange]
    bulb_type: BulbClass
    fw_version: Optional[str]
    white_channels: Optional[int]
    white_to_color_ratio: Optional[int]

    def as_dict(self):
        """Convert to a dict."""
        dict_self = dataclasses.asdict(self)
        dict_self["bulb_type"] = self.bulb_type.name
        return dict_self

    @staticmethod
    def from_data(
        module_name: str,
        kelvin_list: Optional[List[float]],
        fw_version: Optional[str],
        white_channels: Optional[int],
        white_to_color_ratio: Optional[int],
        type_id: Optional[int],
    ) -> "BulbType":
        if module_name:
            try:
                # parse the features from name
                _identifier = module_name.split("_")[1]
            # Throw exception if index can not be found
            except IndexError:
                raise WizLightNotKnownBulb(
                    f"The bulb type could not be determined from the module name: {module_name}"
                )
            if "RGB" in _identifier:  # full RGB bulb
                bulb_type = BulbClass.RGB
                effect = True
            elif "TW" in _identifier:  # Non RGB but tunable white bulb
                bulb_type = BulbClass.TW
                effect = True
            elif "SOCKET" in _identifier:  # A smart socket
                bulb_type = BulbClass.SOCKET
                effect = False
            else:  # Plain brightness-only bulb
                bulb_type = BulbClass.DW
                effect = "DH" in _identifier or "SH" in _identifier
            dual_head = "DH" in _identifier
        elif type_id is not None:
            if type_id not in KNOWN_TYPE_IDS:
                _LOGGER.warning(
                    "Unknown typeId: %s, please report what kind of bulb "
                    "this is at https://github.com/sbidy/pywizlight/issues/new",
                    type_id,
                )
            bulb_type = KNOWN_TYPE_IDS.get(type_id, BulbClass.DW)
            dual_head = False
            effect = True
        else:
            raise WizLightNotKnownBulb(
                f"The bulb type could not be determined from the module name: {module_name} or type_id"
            )

        if kelvin_list:
            kelvin_range: Optional[KelvinRange] = KelvinRange(
                min=int(min(kelvin_list)), max=int(max(kelvin_list))
            )
        elif bulb_type in (BulbClass.RGB, BulbClass.TW):
            raise WizLightNotKnownBulb(
                f"Unable to determine required kelvin range for a {bulb_type.value} device"
            )
        else:
            kelvin_range = None

        features = Features(
            **_BASE_FEATURE_MAP[bulb_type], dual_head=dual_head, effect=effect
        )

        return BulbType(
            bulb_type=bulb_type,
            name=module_name,
            features=features,
            kelvin_range=kelvin_range,
            fw_version=fw_version,
            white_channels=white_channels,
            white_to_color_ratio=white_to_color_ratio,
        )



"""Tests for the Bulb API with a wall switch dimmer."""
from typing import AsyncGenerator

import pytest

from pywizlight import wizlight
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def wall_switch() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP01_DIMTRIACS_01", firmware_version="1.16.68"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_model_description_wall_switch(wall_switch: wizlight) -> None:
    """Test fetching the model description wall switch."""
    bulb_type = await wall_switch.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=False, color_tmp=False, effect=False, brightness=True, dual_head=False
        ),
        name="ESP01_DIMTRIACS_01",
        kelvin_range=KelvinRange(max=2700, min=2700),
        bulb_type=BulbClass.DW,
        fw_version="1.16.68",
        white_channels=1,
        white_to_color_ratio=20,
    )



"""Tests for the Bulb API with a light strip."""
from typing import AsyncGenerator

import pytest

from pywizlight import wizlight
from pywizlight.exceptions import WizLightNotKnownBulb
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def invalid_module_bulb() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="INVALID", firmware_version="1.16.64"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_model_description_no_module_bulb(invalid_module_bulb: wizlight) -> None:
    """Test fetching the model description for a bulb with an invalid module."""
    with pytest.raises(WizLightNotKnownBulb):
        await invalid_module_bulb.get_bulbtype()



"""Tests for the Bulb API with a Squire."""
from typing import AsyncGenerator

import pytest

from pywizlight import PilotBuilder, wizlight
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def squire() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP20_DHRGB_01B", firmware_version="1.21.40"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_setting_ratio(squire: wizlight) -> None:
    """Test setting ratio."""
    await squire.set_ratio(50)
    state = await squire.updateState()
    assert state and state.get_ratio() == 50
    await squire.turn_on(PilotBuilder(ratio=20))
    state = await squire.updateState()
    assert state and state.get_ratio() == 20
    with pytest.raises(ValueError):
        await squire.set_ratio(500)


@pytest.mark.asyncio
async def test_model_description_squire(squire: wizlight) -> None:
    """Test fetching the model description for a squire."""
    bulb_type = await squire.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=True, color_tmp=True, effect=True, brightness=True, dual_head=True
        ),
        name="ESP20_DHRGB_01B",
        kelvin_range=KelvinRange(max=6500, min=2200),
        bulb_type=BulbClass.RGB,
        fw_version="1.21.40",
        white_channels=2,
        white_to_color_ratio=20,
    )



"""Models."""
import dataclasses
from typing import Dict, List


@dataclasses.dataclass(frozen=True)
class DiscoveredBulb:
    """Representation of discovered bulb."""

    ip_address: str
    mac_address: str


@dataclasses.dataclass
class BulbRegistry:
    """Representation of the bulb registry."""

    bulbs_by_mac: Dict[str, DiscoveredBulb] = dataclasses.field(default_factory=dict)

    def register(self, bulb: DiscoveredBulb) -> None:
        """Register a new bulb."""
        self.bulbs_by_mac[bulb.mac_address] = bulb

    def bulbs(self) -> List[DiscoveredBulb]:
        """Get all present bulbs."""
        return list(self.bulbs_by_mac.values())



"""Tests for the Bulb API with a light strip."""
from typing import AsyncGenerator

import pytest

from pywizlight import wizlight
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def dimmable_bulb() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP05_SHDW_21", firmware_version="1.25.0"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_model_description_dimmable_bulb(dimmable_bulb: wizlight) -> None:
    """Test fetching the model description dimmable bulb."""
    bulb_type = await dimmable_bulb.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=False, color_tmp=False, effect=True, brightness=True, dual_head=False
        ),
        name="ESP05_SHDW_21",
        kelvin_range=KelvinRange(max=2700, min=2700),
        bulb_type=BulbClass.DW,
        fw_version="1.25.0",
        white_channels=1,
        white_to_color_ratio=20,
    )


@pytest.mark.asyncio
async def test_supported_scenes(dimmable_bulb: wizlight) -> None:
    """Test supported scenes."""
    assert await dimmable_bulb.getSupportedScenes() == [
        "Wake up",
        "Bedtime",
        "Cool white",
        "Night light",
        "Candlelight",
        "Golden white",
        "Pulse",
        "Steampunk",
    ]



"""Tests for discovery."""
import asyncio
import contextlib
import logging
from typing import AsyncGenerator, Tuple
from unittest.mock import MagicMock, patch

import pytest

from pywizlight.discovery import (
    BroadcastProtocol,
    DiscoveredBulb,
    discover_lights,
    find_wizlights,
)

logging.getLogger("pywizlight").setLevel(logging.DEBUG)


@pytest.fixture
async def mock_discovery_aio_protocol() -> AsyncGenerator:
    """Fixture to mock an asyncio connection."""
    loop = asyncio.get_running_loop()
    future: asyncio.Future[
        Tuple[asyncio.DatagramProtocol, BroadcastProtocol]
    ] = asyncio.Future()

    async def _wait_for_connection():
        transport, protocol = await future
        await asyncio.sleep(0)
        await asyncio.sleep(0)
        return transport, protocol

    async def _mock_create_datagram_endpoint(func, sock=None):
        protocol: BroadcastProtocol = func()
        transport = MagicMock()
        protocol.connection_made(transport)
        with contextlib.suppress(asyncio.InvalidStateError):
            future.set_result((transport, protocol))
        return transport, protocol

    with patch.object(loop, "create_datagram_endpoint", _mock_create_datagram_endpoint):
        yield _wait_for_connection


@pytest.mark.asyncio
async def test_find_wizlights(mock_discovery_aio_protocol):
    """Test find_wizlights."""
    task = asyncio.create_task(
        find_wizlights(wait_time=0.02, broadcast_address="192.168.213.252")
    )
    transport_protocol = await mock_discovery_aio_protocol()
    protocol: BroadcastProtocol = transport_protocol[1]
    protocol.datagram_received(
        b'{"method":"registration","env":"pro","result":{"mac":"d8a01199cf31","success":true}}',
        ("1.3.4.2", 1234),
    )
    protocol.datagram_received(
        b"garbage",
        ("1.3.4.2", 1234),
    )
    protocol.connection_lost(None)
    bulbs = await task
    assert bulbs == [DiscoveredBulb(ip_address="1.3.4.2", mac_address="d8a01199cf31")]


@pytest.mark.asyncio
async def test_discover_lights_fails(mock_discovery_aio_protocol):
    """Test discover_lights."""
    task = asyncio.create_task(discover_lights(wait_time=0.02))
    transport_protocol = await mock_discovery_aio_protocol()
    protocol: BroadcastProtocol = transport_protocol[1]
    protocol.connection_lost(OSError)
    with pytest.raises(OSError):
        await task


@pytest.mark.asyncio
async def test_discover_lights(mock_discovery_aio_protocol):
    """Test discover_lights success."""
    task = asyncio.create_task(discover_lights(wait_time=0.02))
    transport_protocol = await mock_discovery_aio_protocol()
    protocol: BroadcastProtocol = transport_protocol[1]
    protocol.datagram_received(
        b'{"method":"registration","env":"pro","result":{"mac":"d8a01199cf31","success":true}}',
        ("1.3.4.2", 1234),
    )
    protocol.datagram_received(
        b'{"method":"registration","env":"pro","result":{"mac":"d8a01199cf32","success":true}}',
        ("1.3.4.3", 1234),
    )
    protocol.connection_lost(None)
    bulbs = await task
    assert len(bulbs) == 2



"""Tests for the Bulb API with a rgbtw bulb."""
from typing import AsyncGenerator

import pytest

from pywizlight import wizlight
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def rgbw_bulb() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP20_SHRGBC_01", firmware_version="1.21.4"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_model_description_rgbw_bulb(rgbw_bulb: wizlight) -> None:
    """Test fetching the model description rgbw bulb."""
    bulb_type = await rgbw_bulb.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=True, color_tmp=True, effect=True, brightness=True, dual_head=False
        ),
        name="ESP20_SHRGBC_01",
        kelvin_range=KelvinRange(max=6500, min=2200),
        bulb_type=BulbClass.RGB,
        fw_version="1.21.4",
        white_channels=1,
        white_to_color_ratio=30,
    )


@pytest.mark.asyncio
async def test_supported_scenes(rgbw_bulb: wizlight) -> None:
    """Test supported scenes."""
    assert await rgbw_bulb.getSupportedScenes() == [
        "Ocean",
        "Romance",
        "Sunset",
        "Party",
        "Fireplace",
        "Cozy",
        "Forest",
        "Pastel Colors",
        "Wake up",
        "Bedtime",
        "Warm White",
        "Daylight",
        "Cool white",
        "Night light",
        "Focus",
        "Relax",
        "True colors",
        "TV time",
        "Plantgrowth",
        "Spring",
        "Summer",
        "Fall",
        "Deepdive",
        "Jungle",
        "Mojito",
        "Club",
        "Christmas",
        "Halloween",
        "Candlelight",
        "Golden white",
        "Pulse",
        "Steampunk",
        "Rhythm",
    ]



"""Command-line interface to interact with wizlight devices."""
import asyncio
from functools import wraps
from typing import Any, Callable, Coroutine, TypeVar

import click

from pywizlight import PilotBuilder, discovery, wizlight

T = TypeVar("T")


def coro(f: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., T]:
    """Allow to use async in click."""

    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> T:
        """Async wrapper."""
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group()
@click.version_option()
def main() -> None:
    """Command-line tool to interact with Wizlight bulbs."""


@main.command("discover")
@coro
@click.option(
    "--b",
    prompt="Set the broadcast address",
    help="Define the broadcast address like 192.168.1.255.",
)
async def discover(b: str) -> None:
    """Discovery bulb in the local network."""
    click.echo(f"Search for bulbs in {b} ... ")

    bulbs = await discovery.find_wizlights(broadcast_address=b)
    for bulb in bulbs:
        click.echo(bulb.__dict__)


@main.command("on")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
@click.option(
    "--k",
    prompt="Kelvin for temperature.",
    help="Kelvin value (1000-8000) for turn on. Default 3000",
    default=3000,
    type=int,
)
@click.option(
    "--brightness",
    prompt="Set the brightness value 0-255",
    help="Brightness for turn on. Default 128",
    default=128,
    type=int,
)
async def turn_on(ip: str, k: int, brightness: int) -> None:
    """Turn a given bulb on."""
    click.echo(f"Turning on {ip}")
    bulb = wizlight(ip)
    if bulb and 1000 <= k <= 6800 and 0 <= brightness <= 255:
        await bulb.turn_on(PilotBuilder(colortemp=k, brightness=brightness))
    else:
        click.echo("Error - values are not correct. Type --help for help.")
    await bulb.async_close()


@main.command("set-state")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
@click.option(
    "--k",
    prompt="Kelvin for temperature.",
    help="Kelvin value (1000-8000) for turn on. Default 3000",
    default=3000,
)
@click.option(
    "--brightness",
    prompt="Set the brightness value 0-255",
    help="Brightness for turn on. Default 128",
    default=128,
)
async def set_state(ip: str, k: int, brightness: int) -> None:
    """Set the current state of a given bulb."""
    click.echo(f"Turning on {ip}")
    bulb = wizlight(ip)
    if bulb and 1000 <= k <= 6800 and 0 <= brightness <= 255:
        await bulb.set_state(PilotBuilder(colortemp=k, brightness=brightness))
    else:
        click.echo("Error - values are not correct. Type --help for help.")
    await bulb.async_close()


@main.command("off")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
async def turn_off(ip: str) -> None:
    """Turn a given bulb off."""
    click.echo(f"Turning off {ip}")
    bulb = wizlight(ip)
    await bulb.turn_off()
    await bulb.async_close()


@main.command("state")
@coro
@click.option("--ip", prompt="IP address of the bulb", help="IP address of the bulb.")
async def state(ip: str) -> None:
    """Get the current state of a given bulb."""
    click.echo(f"Get the state from {ip}")
    bulb = wizlight(ip)
    bulb_state = await bulb.updateState()
    if bulb_state:
        click.echo(bulb_state.pilotResult)
    else:
        click.echo("Did not get state from bulb")
    await bulb.async_close()


if __name__ == "__main__":
    main()



"""Tests for the Bulb API with a light strip."""
from typing import AsyncGenerator

import pytest

from pywizlight import wizlight
from pywizlight.exceptions import WizLightConnectionError
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def broken_bulb() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="BROKEN_JSON", firmware_version="1.0.0"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.mark.asyncio
async def test_model_description_broken_bulb(
    broken_bulb: wizlight, caplog: pytest.LogCaptureFixture
) -> None:
    """Test fetching the model description for an bulb that generates broken json."""
    with pytest.raises(WizLightConnectionError):
        await broken_bulb.get_bulbtype()
    assert "Failed to decode message" in caplog.text



"""Tests for the Bulb API."""
from typing import AsyncGenerator
from unittest.mock import patch

import pytest

from pywizlight import SCENES, PilotBuilder, wizlight
from pywizlight.bulb import states_match
from pywizlight.bulblibrary import BulbClass, BulbType, Features, KelvinRange
from pywizlight.discovery import discover_lights
from pywizlight.exceptions import WizLightTimeOutError
from pywizlight.tests.fake_bulb import startup_bulb


@pytest.fixture()
async def correct_bulb() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP01_SHRGB_03", firmware_version="1.25.0"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.fixture()
async def power_socket_old_firmware() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP25_SOCKET_01", firmware_version="1.26.1"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.fixture()
async def power_socket() -> AsyncGenerator[wizlight, None]:
    shutdown, port = await startup_bulb(
        module_name="ESP25_SOCKET_01", firmware_version="1.26.2"
    )
    bulb = wizlight(ip="127.0.0.1", port=port)
    yield bulb
    await bulb.async_close()
    shutdown()


@pytest.fixture()
async def bad_bulb() -> AsyncGenerator[wizlight, None]:
    bulb = wizlight(ip="1.1.1.1")
    yield bulb
    await bulb.async_close()


# Non-Error states - PilotBuilder - Turn On
@pytest.mark.asyncio
async def test_Bulb_Discovery(correct_bulb: wizlight) -> None:
    """Test discovery function."""
    # Use a random available port since otherwise the
    # test may fail
    with patch("pywizlight.discovery.PORT", 0):
        bulbs = await discover_lights(broadcast_space="192.168.178.255", wait_time=0.02)
    for bulb in bulbs:
        with patch("pywizlight.bulb.FIRST_SEND_INTERVAL", 0.01), patch(
            "pywizlight.bulb.TIMEOUT", 0.01
        ):
            state = await bulb.updateState()
        assert state and state.get_state() is False


@pytest.mark.asyncio
async def test_PilotBuilder_state(correct_bulb: wizlight) -> None:
    """Test State."""
    state = await correct_bulb.updateState()
    assert state and state.get_state() is False


@pytest.mark.asyncio
async def test_PilotBuilder_colortemp(correct_bulb: wizlight) -> None:
    """Test Color Temp."""
    await correct_bulb.turn_on(PilotBuilder(colortemp=2800))
    state = await correct_bulb.updateState()

    assert state and state.get_colortemp() == 2800


@pytest.mark.asyncio
async def test_PilotBuilder_brightness(correct_bulb: wizlight) -> None:
    """Test Brightness."""
    await correct_bulb.turn_on(PilotBuilder(brightness=10))
    state = await correct_bulb.updateState()
    # 10% == 26 in Hex
    assert state and state.get_brightness() == 26


@pytest.mark.asyncio
async def test_PilotBuilder_warm_wite(correct_bulb: wizlight) -> None:
    """Test Warm White."""
    await correct_bulb.turn_on(PilotBuilder(warm_white=255))
    state = await correct_bulb.updateState()

    assert state and state.get_warm_white() == 255


@pytest.mark.asyncio
async def test_PilotBuilder_cold_white(correct_bulb: wizlight) -> None:
    """Test Cold White."""
    await correct_bulb.turn_on(PilotBuilder(cold_white=255))
    state = await correct_bulb.updateState()

    assert state and state.get_cold_white() == 255


@pytest.mark.asyncio
async def test_PilotBuilder_rgb(correct_bulb: wizlight) -> None:
    """Test RGB Value."""
    await correct_bulb.turn_on(PilotBuilder(rgb=(0, 128, 255)))
    state = await correct_bulb.updateState()

    assert state and state.get_rgb() == (0, 127, 255)


@pytest.mark.asyncio
async def test_PilotBuilder_hucolor(correct_bulb: wizlight) -> None:
    """Test RGB Value via hucolor."""
    await correct_bulb.turn_on(PilotBuilder(hucolor=(100, 50)))
    state = await correct_bulb.updateState()

    assert state and state.get_rgb() == (88.0, 255.0, 0.0)


@pytest.mark.asyncio
async def test_setting_rgbw(correct_bulb: wizlight) -> None:
    """Test setting rgbw."""
    await correct_bulb.turn_on(PilotBuilder(rgbw=(1, 2, 3, 4)))
    state = await correct_bulb.updateState()
    assert state and state.get_rgbw() == (1, 2, 3, 4)


@pytest.mark.asyncio
async def test_PilotBuilder_scene(correct_bulb: wizlight) -> None:
    """Test scene."""
    await correct_bulb.turn_on(PilotBuilder(scene=1))
    state = await correct_bulb.updateState()

    assert state and state.get_scene() == SCENES[1]
    state.pilotResult["schdPsetId"] = True
    assert state.get_scene() == SCENES[1000]


@pytest.mark.asyncio
async def test_PilotBuilder_scene_empty(correct_bulb: wizlight) -> None:
    """Test scene with no scene set."""
    state = await correct_bulb.updateState()
    assert state is not None
    if "sceneId" in state.pilotResult:
        del state.pilotResult["sceneId"]
    assert state and state.get_scene() is None


@pytest.mark.asyncio
async def test_PilotBuilder_speed(correct_bulb: wizlight) -> None:
    """Test speed."""
    await correct_bulb.turn_on(PilotBuilder(scene=1, speed=50))
    state = await correct_bulb.updateState()

    assert state and state.get_scene() == SCENES[1]
    assert state and state.get_speed() == 50


@pytest.mark.asyncio
async def test_set_speed(correct_bulb: wizlight) -> None:
    """Set speed."""
    await correct_bulb.set_speed(125)
    state = await correct_bulb.updateState()
    assert state and state.get_speed() == 125


@pytest.mark.asyncio
async def test_get_source(correct_bulb: wizlight) -> None:
    """Test getting the source."""
    state = await correct_bulb.updateState()
    assert state and state.get_source() == "udp"


# ------ Error states -------------------------------------
@pytest.mark.asyncio
async def test_error_PilotBuilder_brightness(correct_bulb: wizlight) -> None:
    """Error Brightness."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(brightness=500))


@pytest.mark.asyncio
async def test_error_PilotBuilder_warm_wite(correct_bulb: wizlight) -> None:
    """Error Warm White."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(warm_white=300))
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(warm_white=-1))


@pytest.mark.asyncio
async def test_error_PilotBuilder_cold_white_upper(correct_bulb: wizlight) -> None:
    """Error Cold White."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(cold_white=300))


@pytest.mark.asyncio
async def test_error_PilotBuilder_cold_white_lower(correct_bulb: wizlight) -> None:
    """Error Cold White."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(cold_white=-1))


@pytest.mark.asyncio
async def test_error_PilotBuilder_r(correct_bulb: wizlight) -> None:
    """Error Red Value."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(rgb=(300, 0, 0)))


@pytest.mark.asyncio
async def test_error_PilotBuilder_green(correct_bulb: wizlight) -> None:
    """Error Green Value."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(rgb=(0, 300, 0)))


@pytest.mark.asyncio
async def test_error_PilotBuilder_blue(correct_bulb: wizlight) -> None:
    """Error Blue Value."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(rgb=(0, 0, 300)))


@pytest.mark.asyncio
async def test_error_PilotBuilder_cold_white(correct_bulb: wizlight) -> None:
    """Error Cold White Value."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(cold_white=9999))


@pytest.mark.asyncio
async def test_error_PilotBuilder_scene(correct_bulb: wizlight) -> None:
    """Error scene."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(scene=532))


@pytest.mark.asyncio
async def test_error_PilotBuilder_speed(correct_bulb: wizlight) -> None:
    """Error speed."""
    with pytest.raises(ValueError):
        await correct_bulb.turn_on(PilotBuilder(speed=532))


@pytest.mark.asyncio
async def test_error_set_speed(correct_bulb: wizlight) -> None:
    """Error speed."""
    with pytest.raises(ValueError):
        await correct_bulb.set_speed(532)


@pytest.mark.asyncio
async def test_fw_version(correct_bulb: wizlight) -> None:
    """Test fetching the firmware version."""
    bulb_type = await correct_bulb.get_bulbtype()
    assert bulb_type == BulbType(
        features=Features(
            color=True, color_tmp=True, effect=True, brightness=True, dual_head=False
        ),
        name="ESP01_SHRGB_03",
        kelvin_range=KelvinRange(max=6500, min=2200),
        bulb_type=BulbClass.RGB,
        fw_version="1.25.0",
        white_channels=1,
        white_to_color_ratio=30,
    )
    assert correct_bulb.mac == "a8bb5006033d"


@pytest.mark.asyncio
async def test_get_mac(correct_bulb: wizlight) -> None:
    """Test getting the mac address."""
    mac = await correct_bulb.getMac()
    assert mac == "a8bb5006033d"
    mac = await correct_bulb.getMac()
    assert mac == "a8bb5006033d"


@pytest.mark.asyncio
async def test_get_power(power_socket: wizlight) -> None:
    """Test getting power in watts."""
    watts = await power_socket.get_power()
    assert watts == 1065.385


@pytest.mark.asyncio
async def test_get_power_old_firmware(power_socket_old_firmware: wizlight) -> None:
    """Test getting power in watts."""
    watts = await power_socket_old_firmware.get_power()
    assert watts is None


@pytest.mark.asyncio
async def test_get_power_unsupported_device(correct_bulb: wizlight) -> None:
    """Test getting power in watts."""
    assert await correct_bulb.get_power() is None
    assert await correct_bulb.get_power() is None


# Error states / Timout
@pytest.mark.asyncio
async def test_timeout(bad_bulb: wizlight) -> None:
    """Test the timout exception after."""
    with pytest.raises(WizLightTimeOutError), patch(
        "pywizlight.bulb.FIRST_SEND_INTERVAL", 0.01
    ), patch("pywizlight.bulb.TIMEOUT", 0.01):
        await bad_bulb.getBulbConfig()


@pytest.mark.asyncio
async def test_timeout_PilotBuilder(bad_bulb: wizlight) -> None:
    """Test Timout for Result."""
    # check if the bulb state it given as bool - mock ?
    with pytest.raises(WizLightTimeOutError), patch(
        "pywizlight.bulb.FIRST_SEND_INTERVAL", 0.01
    ), patch("pywizlight.bulb.TIMEOUT", 0.01):
        await bad_bulb.turn_on(PilotBuilder(brightness=255))


@pytest.mark.asyncio
async def test_states_match_with_occupancy() -> None:
    """Test states match always sends pir updates but we ignore mqttCd, rssi, and ts."""
    state_off_ios = {
        "mac": "a8bb50d46a1c",
        "rssi": -70,
        "src": "ios",
        "mqttCd": 0,
        "ts": 1644440635,
        "state": False,
        "sceneId": 0,
    }
    state_on_ios = {
        "mac": "a8bb50d46a1c",
        "rssi": -45,
        "src": "ios",
        "mqttCd": 0,
        "ts": 1644440662,
        "state": False,
        "sceneId": 27,
        "speed": 100,
        "dimming": 100,
    }
    state_on_hb = {
        "mac": "a8bb50d46a1c",
        "rssi": -45,
        "src": "hb",
        "mqttCd": 0,
        "ts": 1644440642,
        "state": False,
        "sceneId": 27,
        "speed": 100,
        "dimming": 100,
    }
    ios_scene27 = {
        "mac": "a8bb50d46a1c",
        "rssi": -48,
        "src": "ios",
        "state": True,
        "sceneId": 27,
        "speed": 100,
        "dimming": 100,
    }
    ios_off = {
        "mac": "a8bb50d46a1c",
        "rssi": -69,
        "src": "ios",
        "state": False,
        "sceneId": 0,
    }
    occupancy_detected_scene27 = {
        "mac": "a8bb50d46a1c",
        "rssi": -48,
        "src": "pir",
        "state": True,
        "sceneId": 27,
        "speed": 100,
        "dimming": 100,
    }
    occupancy_not_detected = {
        "mac": "a8bb50d46a1c",
        "rssi": -69,
        "src": "pir",
        "state": False,
        "sceneId": 0,
    }

    assert states_match(state_off_ios, state_off_ios)
    assert not states_match(state_off_ios, state_on_ios)
    assert states_match(
        state_on_ios, state_on_hb
    )  # source change does not matter unless its a PIR
    assert not states_match(
        ios_scene27, occupancy_detected_scene27
    )  # source change matters since its a PIR
    assert states_match(occupancy_detected_scene27, occupancy_detected_scene27)
    assert not states_match(
        ios_off, occupancy_not_detected
    )  # source change matters since its a PIR


@pytest.mark.asyncio
async def test_states_match_with_button() -> None:
    """Test states match always sends button updates but we ignore mqttCd, rssi, and ts."""

    button_on_press = {
        "mac": "a8bb50d46a1c",
        "rssi": -48,
        "src": "wfa1",
        "state": True,
    }
    button_off_press = {
        "mac": "a8bb50d46a1c",
        "rssi": -69,
        "src": "wfa2",
        "state": False,
        "sceneId": 0,
    }
    button_1_press_state_false = {
        "mac": "a8bb50d46a1c",
        "rssi": -69,
        "src": "wfa16",
        "state": False,
        "sceneId": 0,
    }
    button_1_press_state_true = {
        "mac": "a8bb50d46a1c",
        "rssi": -69,
        "src": "wfa16",
        "state": True,
        "sceneId": 0,
    }
    ios_off = {
        "mac": "a8bb50d46a1c",
        "rssi": -69,
        "src": "ios",
        "state": False,
        "sceneId": 0,
    }
    assert not states_match(button_on_press, button_off_press)
    assert not states_match(button_1_press_state_false, button_1_press_state_true)
    assert not states_match(
        ios_off, button_1_press_state_false
    )  # source change matters since its a button



"""Definition of the supported bulb effects."""

from typing import Dict, Iterable, List, cast

from pywizlight.bulblibrary import BulbClass

SCENES = {
    1: "Ocean",
    2: "Romance",
    3: "Sunset",
    4: "Party",
    5: "Fireplace",
    6: "Cozy",
    7: "Forest",
    8: "Pastel Colors",
    9: "Wake up",
    10: "Bedtime",
    11: "Warm White",
    12: "Daylight",
    13: "Cool white",
    14: "Night light",
    15: "Focus",
    16: "Relax",
    17: "True colors",
    18: "TV time",
    19: "Plantgrowth",
    20: "Spring",
    21: "Summer",
    22: "Fall",
    23: "Deepdive",
    24: "Jungle",
    25: "Mojito",
    26: "Club",
    27: "Christmas",
    28: "Halloween",
    29: "Candlelight",
    30: "Golden white",
    31: "Pulse",
    32: "Steampunk",
    1000: "Rhythm",
}
SCENE_NAME_TO_ID = {scene_name: scene_id for (scene_id, scene_name) in SCENES.items()}
TW_SCENES = [6, 9, 10, 11, 12, 13, 14, 15, 16, 18, 29, 30, 31, 32]
DW_SCENES = [9, 10, 13, 14, 29, 30, 31, 32]

SCENES_BY_CLASS: Dict[BulbClass, List[str]] = {
    BulbClass.RGB: list(cast(Iterable, SCENES.values())),
    BulbClass.TW: [SCENES[key] for key in TW_SCENES],
    BulbClass.DW: [SCENES[key] for key in DW_SCENES],
}


def get_id_from_scene_name(scene: str) -> int:
    """Return the id of an given scene name.

    :param scene: Name of the scene
    :raises ValueError: Return if not in scene list
    :return: ID of the scene
    """
    scene_id = SCENE_NAME_TO_ID.get(scene)
    if not scene_id:
        raise ValueError(f"Scene '{scene}' not in scene list.")
    return scene_id
